#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module holds internally-used functions to determine various properties
of a glyph.

These properties assist in applying automatisms to glyphs when round-
tripping.
"""


import collections
import re
from fontTools import unicodedata
import xml.etree.ElementTree

import fontTools.agl


__all__ = ["get_glyph", "GlyphData"]

# This is an internally-used named tuple and not meant to be a GSGlyphData replacement.
Glyph = collections.namedtuple(
    "Glyph",
    "name, production_name, unicode, category, subCategory, script, description",
)

# Global variable holding the actual GlyphData data, assigned on first use.
GLYPHDATA = None


class GlyphData:
    """Map (alternative) names and production names to GlyphData data.

    This class holds the GlyphData data as provided on
    https://github.com/schriftgestalt/GlyphsInfo and provides lookup by
    name, alternative name and production name through normal
    dictionaries.
    """

    __slots__ = ["names", "alternative_names", "production_names", "unicodes"]

    def __init__(
        self, name_mapping, alt_name_mapping, production_name_mapping, unicodes_mapping
    ):
        self.names = name_mapping
        self.alternative_names = alt_name_mapping
        self.production_names = production_name_mapping
        self.unicodes = unicodes_mapping

    @classmethod
    def from_files(cls, *glyphdata_files):
        """Return GlyphData holding data from a list of XML file paths."""
        name_mapping = {}
        alt_name_mapping = {}
        production_name_mapping = {}
        unicodes_mapping = {}

        for glyphdata_file in glyphdata_files:
            glyph_data = xml.etree.ElementTree.parse(glyphdata_file).getroot()
            for glyph in glyph_data:
                glyph_name = glyph.attrib["name"]
                glyph_name_alternatives = glyph.attrib.get("altNames")
                glyph_name_production = glyph.attrib.get("production")
                glyph_unicode = glyph.attrib.get("unicode")

                name_mapping[glyph_name] = glyph.attrib
                if glyph_name_alternatives:
                    alternatives = glyph_name_alternatives.replace(" ", "").split(",")
                    for glyph_name_alternative in alternatives:
                        alt_name_mapping[glyph_name_alternative] = glyph.attrib
                if glyph_name_production:
                    production_name_mapping[glyph_name_production] = glyph.attrib
                if glyph_unicode:
                    unicodes_mapping[glyph_unicode] = glyph.attrib

        return cls(
            name_mapping, alt_name_mapping, production_name_mapping, unicodes_mapping
        )


def get_glyph(glyph_name, data=None, unicodes=None):
    """Return a named tuple (Glyph) containing information derived from a glyph
    name akin to GSGlyphInfo.

    The information is derived from an included copy of GlyphData.xml
    and GlyphData_Ideographs.xml, going by the glyph name or unicode fallback.
    """

    # Read data on first use.
    global GLYPHDATA
    if GLYPHDATA is None:
        try:
            from importlib.resources import files
        except ImportError:
            # Python <= 3.8 backport
            from importlib_resources import files

        data_dir = files("glyphsLib.data")
        with (data_dir / "GlyphData.xml").open("rb") as f1:
            with (data_dir / "GlyphData_Ideographs.xml").open("rb") as f2:
                GLYPHDATA = GlyphData.from_files(f1, f2)
                assert len(GLYPHDATA.names) > 30000

    if data is None:
        data = GLYPHDATA

    # Look up data by full glyph name first.
    attributes = _lookup_attributes(glyph_name, data)

    # Look up by unicode
    if attributes == {} and unicodes is not None:
        for unicode in unicodes:
            attributes = _lookup_attributes_by_unicode(unicode, data)
            if attributes:
                break

    production_name = attributes.get("production")
    if production_name is None:
        production_name = _construct_production_name(glyph_name, data=data)

    unicode_value = attributes.get("unicode")

    category = attributes.get("category")
    sub_category = attributes.get("subCategory")
    if category is None:
        category, sub_category = _construct_category(glyph_name, data)

    # TODO: Determine script in ligatures.
    script = attributes.get("script")
    description = attributes.get("description")

    return Glyph(
        glyph_name,
        production_name,
        unicode_value,
        category,
        sub_category,
        script,
        description,
    )


def _lookup_attributes(glyph_name, data):
    """Look up glyph attributes in data by glyph name, alternative name or
    production name in order or return empty dictionary.

    Look up by alternative and production names for legacy projects and
    because of issue #232.
    """
    attributes = (
        data.names.get(glyph_name)
        or data.alternative_names.get(glyph_name)
        or data.production_names.get(glyph_name)
        or {}
    )
    # If we are using custom GlyphData, fallback to default GlyphData
    if not attributes and data is not GLYPHDATA:
        attributes = _lookup_attributes(glyph_name, GLYPHDATA)
    return attributes


def _lookup_attributes_by_unicode(unicode, data):
    """Look up glyph attributes in data by unicode
    or return empty dictionary.
    """
    attributes = data.unicodes.get(unicode) or {}
    return attributes


def _agl_compliant_name(glyph_name):
    """Return an AGL-compliant name string or None if we can't make one."""
    MAX_GLYPH_NAME_LENGTH = 63
    clean_name = re.sub("[^0-9a-zA-Z_.]", "", glyph_name)
    if len(clean_name) > MAX_GLYPH_NAME_LENGTH:
        return None
    return clean_name


def _is_unicode_u_value(name):
    """Return whether we are looking at a uXXXX value."""
    return name.startswith("u") and all(
        part_char in "0123456789ABCDEF" for part_char in name[1:]
    )


def _construct_category(glyph_name, data):
    """Derive (sub)category of a glyph name."""
    # Glyphs creates glyphs that start with an underscore as "non-exportable" glyphs or
    # construction helpers without a category.
    if glyph_name.startswith("_"):
        return None, None

    # Glyph variants (e.g. "fi.alt") don't have their own entry, so we strip e.g. the
    # ".alt" and try a second lookup with just the base name. A variant is hopefully in
    # the same category as its base glyph.
    base_name = _split_glyph_name(glyph_name, data)[0]
    base_attribute = _lookup_attributes(base_name, data) or {}
    if base_attribute:
        category = base_attribute.get("category")
        sub_category = base_attribute.get("subCategory")
        return category, sub_category

    # Detect ligatures.
    if "_" in base_name:
        base_names = _split_ligature_glyph_name(base_name, data)
        base_names_attributes = [_lookup_attributes(name, data) for name in base_names]
        first_attribute = base_names_attributes[0]

        # If the first part is a Mark, Glyphs 2.6 declares the entire glyph a Mark
        if first_attribute.get("category") == "Mark":
            category = first_attribute.get("category")
            sub_category = first_attribute.get("subCategory")
            return category, sub_category

        # If the first part is a Letter...
        if first_attribute.get("category") == "Letter":
            # ... and the rest are only marks or separators or don't exist, the
            # sub_category is that of the first part ...
            if all(
                a.get("category") in (None, "Mark", "Separator")
                for a in base_names_attributes[1:]
            ):
                category = first_attribute.get("category")
                sub_category = first_attribute.get("subCategory")
                return category, sub_category
            # ... otherwise, a ligature.
            category = first_attribute.get("category")
            sub_category = "Ligature"
            return category, sub_category

        # TODO: Cover more cases. E.g. "one_one" -> ("Number", "Ligature") but
        # "one_onee" -> ("Number", "Composition").

    # Still nothing? Maybe we're looking at something like "uni1234.alt", try
    # using fontTools' AGL module to convert the base name to something meaningful.
    # Corner case: when looking at ligatures, names that don't exist in the AGLFN
    # are skipped, so len("acutecomb_o") == 2 but len("dotaccentcomb_o") == 1.
    character = fontTools.agl.toUnicode(base_name)
    if character:
        category, sub_category = _translate_category(
            glyph_name, unicodedata.category(character[0])
        )
        return category, sub_category

    return None, None


def _translate_category(glyph_name, unicode_category):
    """Return a translation from Unicode category letters to Glyphs
    categories."""
    DEFAULT_CATEGORIES = {
        None: ("Letter", None),
        "Cc": ("Separator", None),
        "Cf": ("Separator", "Format"),
        "Cn": ("Symbol", None),
        "Co": ("Letter", "Compatibility"),
        "Ll": ("Letter", None),
        "Lm": ("Letter", "Modifier"),
        "Lo": ("Letter", None),
        "Lt": ("Letter", None),
        "Lu": ("Letter", None),
        "Mc": ("Mark", "Spacing Combining"),
        "Me": ("Mark", "Enclosing"),
        "Mn": ("Mark", "Nonspacing"),
        "Nd": ("Number", "Decimal Digit"),
        "Nl": ("Number", None),
        "No": ("Number", "Decimal Digit"),
        "Pc": ("Punctuation", None),
        "Pd": ("Punctuation", "Dash"),
        "Pe": ("Punctuation", "Parenthesis"),
        "Pf": ("Punctuation", "Quote"),
        "Pi": ("Punctuation", "Quote"),
        "Po": ("Punctuation", None),
        "Ps": ("Punctuation", "Parenthesis"),
        "Sc": ("Symbol", "Currency"),
        "Sk": ("Mark", "Spacing"),
        "Sm": ("Symbol", "Math"),
        "So": ("Symbol", None),
        "Zl": ("Separator", None),
        "Zp": ("Separator", None),
        "Zs": ("Separator", "Space"),
    }

    glyphs_category = DEFAULT_CATEGORIES.get(unicode_category, ("Letter", None))

    # Exception: Something like "one_two" should be a (_, Ligature),
    # "acutecomb_brevecomb" should however stay (Mark, Nonspacing).
    if "_" in glyph_name and glyphs_category[0] != "Mark":
        return glyphs_category[0], "Ligature"

    return glyphs_category


def _split_ligature_glyph_name(name, data):
    # Split name to ligature parts
    parts = name.split("_")

    # If the last part has a script suffix, strip it and re-split the name.
    if "-" in parts[-1]:
        base, script = name.rsplit("-", 1)
        parts = base.split("_")

        # If there is more than one part, try adding the script suffix to each
        # part, if this results in a known glyph name, use it as the part name.
        if len(parts) > 1:
            for i, part in enumerate(parts):
                new = f"{part}-{script}"
                # If the part already has a script suffix, keep it unchanged.
                if "-" in part:
                    continue
                # If the non suffixed name exists and the suffixed name does
                # not exist, keep the part name unchanged.
                if _lookup_attributes(part, data) and not _lookup_attributes(new, data):
                    continue
                parts[i] = new
    else:
        parts = name.split("_")
    return parts


def _split_glyph_name(name, data):
    # Split glyph name into base and suffix
    base, dot, suffix = name.partition(".")

    # If there are more than one suffix (e.g. ".below.ro"), try adding each
    # suffix to the base name, if it results in a known glyph name, use that as
    # base name.
    if dot and dot in suffix:
        suffixes = suffix.split(dot)
        new = base
        while suffixes:
            new += dot + suffixes.pop(0)
            if _lookup_attributes(new, data):
                return new, dot, dot.join(suffixes)
    return base, dot, suffix


def _construct_production_name(glyph_name, data=None):
    """Return the production name for a glyph name from the GlyphData.xml
    database according to the AGL specification.

    This should be run only if there is no official entry with a production
    name in it.

    Handles single glyphs (e.g. "brevecomb") and ligatures (e.g.
    "brevecomb_acutecomb"). Returns None when a valid and semantically
    meaningful production name can't be constructed or when the AGL
    specification would be violated, get_glyph() will use the bare glyph
    name then.

    Note:
    - Glyph name is the full name, e.g. "brevecomb_acutecomb.case".
    - Base name is the base part, e.g. "brevecomb_acutecomb"
    - Suffix is e.g. "case".
    """

    # At this point, we have already checked the data for the full glyph name, so
    # directly go to the base name here (e.g. when looking at "fi.alt").
    base_name, dot, suffix = _split_glyph_name(glyph_name, data)
    glyphinfo = _lookup_attributes(base_name, data)
    if glyphinfo and glyphinfo.get("production"):
        # Found the base glyph.
        return glyphinfo["production"] + dot + suffix

    if glyph_name in fontTools.agl.AGL2UV or base_name in fontTools.agl.AGL2UV:
        # Glyph name is actually an AGLFN name.
        return glyph_name

    if "_" not in base_name:
        # Nothing found so far and the glyph name isn't a ligature ("_"
        # somewhere in it). The name does not carry any discernable Unicode
        # semantics, so just return something sanitized.
        return _agl_compliant_name(glyph_name)

    # So we have a ligature that is not mapped in the data. Split it up and
    # look up the individual parts.
    base_name_parts = _split_ligature_glyph_name(base_name, data)

    # If all parts are in the AGLFN list, the glyph name is our production
    # name already.
    if all(part in fontTools.agl.AGL2UV for part in base_name_parts):
        production_name = "_".join(base_name_parts) + dot + suffix
        return _agl_compliant_name(production_name)

    # Turn all parts of the ligature into production names.
    _character_outside_BMP = False
    production_names = []
    for part in base_name_parts:
        if part in fontTools.agl.AGL2UV:
            # A name present in the AGLFN is a production name already.
            production_names.append(part)
        else:
            part_entry = _lookup_attributes(part, data) or {}
            part_production_name = part_entry.get("production")
            if part_production_name:
                production_names.append(part_production_name)

                # Take note if there are any characters outside the Unicode
                # BMP, e.g. "u10FFF" or "u10FFFF". Do not catch e.g. "u013B"
                # though.
                if len(part_production_name) > 5 and _is_unicode_u_value(
                    part_production_name
                ):
                    _character_outside_BMP = True
            else:
                # We hit a part that does not seem to be a valid glyph name known to us,
                # so the entire glyph name can't carry Unicode meaning. Return it
                # sanitized.
                return _agl_compliant_name(glyph_name)

    # Some names Glyphs uses resolve to other names that are not uniXXXX names and may
    # contain dots (e.g. idotaccent -> i.loclTRK). If there is any name with a "." in
    # it before the last element, punt. We'd have to introduce a "." into the ligature
    # midway, which is invalid according to the AGL. Example: "a_i.loclTRK" is valid,
    # but "a_i.loclTRK_a" isn't.
    if any("." in part for part in production_names[:-1]):
        return _agl_compliant_name(glyph_name)

    # If any production name starts with a "uni" and there are none of the
    # "uXXXXX" format, try to turn all parts into "uni" names and concatenate
    # them.
    if not _character_outside_BMP and any(
        part.startswith("uni") for part in production_names
    ):
        uni_names = []

        for part in production_names:
            if part.startswith("uni"):
                uni_names.append(part[3:])
            elif len(part) == 5 and _is_unicode_u_value(part):
                uni_names.append(part[1:])
            elif part in fontTools.agl.AGL2UV:
                uni_names.append("{:04X}".format(fontTools.agl.AGL2UV[part]))
            else:
                return None

        final_production_name = "uni" + "".join(uni_names) + dot + suffix
    else:
        final_production_name = "_".join(production_names) + dot + suffix

    return _agl_compliant_name(final_production_name)



from types import MappingProxyType
from typing import NamedTuple

from .align_alternate_layers import align_alternate_layers
from .propagate_anchors import propagate_all_anchors

TRANSFORMATIONS = [
    align_alternate_layers,
    propagate_all_anchors,
]


class _CustomParameter(NamedTuple):
    name: str
    default: bool


TRANSFORMATION_CUSTOM_PARAMS = MappingProxyType(
    {
        propagate_all_anchors: _CustomParameter("Propagate Anchors", True),
    }
)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import deque

from .constants import GLYPHS_PREFIX


def to_ufo_names(self, ufo, master, family_name):
    if not self.minimal:
        to_ufo_names_roundtrip(master, ufo)

    is_italic = bool(master.italicAngle)

    styleName = master.name
    ufo.info.familyName = family_name
    ufo.info.styleName = styleName

    # FIXME: (jany) should be the responsibility of ufo2ft?
    # Anyway, only generate the styleMap names if we're not round-tripping
    # (i.e. generating UFOs for fontmake, the traditional use-case of
    # glyphsLib.)
    if not self.minimize_glyphs_diffs:
        styleMapFamilyName, styleMapStyleName = build_stylemap_names(
            family_name=family_name,
            style_name=styleName,
            is_bold=(styleName in ("Bold", "Bold Italic", "Bold Oblique")),
            is_italic=is_italic,
        )
        ufo.info.styleMapFamilyName = styleMapFamilyName
        ufo.info.styleMapStyleName = styleMapStyleName


def to_ufo_names_roundtrip(master, ufo):
    width = master.width
    weight = master.weight
    custom = master.customName

    if weight:
        ufo.lib[GLYPHS_PREFIX + "weight"] = weight
    if width:
        ufo.lib[GLYPHS_PREFIX + "width"] = width
    if custom:
        ufo.lib[GLYPHS_PREFIX + "customName"] = master.customName


def build_stylemap_names(
    family_name, style_name, is_bold=False, is_italic=False, linked_style=None
):
    """Build UFO `styleMapFamilyName` and `styleMapStyleName` based on the
    family and style names, and the entries in the "Style Linking" section
    of the "Instances" tab in the "Font Info".

    The value of `styleMapStyleName` can be either "regular", "bold", "italic"
    or "bold italic", depending on the values of `is_bold` and `is_italic`.

    The `styleMapFamilyName` is a combination of the `family_name` and the
    `linked_style`.

    If `linked_style` is unset or set to 'Regular', the linked style is equal
    to the style_name with the last occurrences of the strings 'Regular',
    'Bold' and 'Italic' stripped from it.
    """

    styleMapStyleName = (
        " ".join(
            s for s in ("bold" if is_bold else "", "italic" if is_italic else "") if s
        )
        or "regular"
    )
    if not linked_style or linked_style == "Regular":
        linked_style = _get_linked_style(style_name, is_bold, is_italic)
    if linked_style:
        styleMapFamilyName = (family_name or "") + " " + linked_style
    else:
        styleMapFamilyName = family_name
    return styleMapFamilyName, styleMapStyleName


def _get_linked_style(style_name, is_bold, is_italic):
    # strip last occurrence of 'Regular', 'Bold', 'Italic' from style_name
    # depending on the values of is_bold and is_italic
    linked_style = deque()
    is_regular = not (is_bold or is_italic)
    for part in reversed(style_name.split()):
        if part == "Regular" and is_regular:
            is_regular = False
        elif part == "Bold" and is_bold:
            is_bold = False
        elif (part == "Italic" or part == "Oblique") and is_italic:
            is_italic = False
        else:
            linked_style.appendleft(part)
    return " ".join(linked_style)


def to_glyphs_family_names(self, ufo, merge=False):
    if not merge:
        # First UFO
        self.font.familyName = ufo.info.familyName
    else:
        # Subsequent UFOs
        if self.font.familyName != ufo.info.familyName:
            raise RuntimeError("All UFOs should have the same family name.")


def to_glyphs_master_names(self, ufo, master):
    name = ufo.info.styleName
    weight = ufo.lib.get(GLYPHS_PREFIX + "weight")
    width = ufo.lib.get(GLYPHS_PREFIX + "width")
    custom = ufo.lib.get(GLYPHS_PREFIX + "customName")

    master.set_all_name_components(name, weight, width, custom)



from __future__ import annotations

from pathlib import Path

import pytest

from glyphsLib import GSFont
from glyphsLib.builder.transformations.align_alternate_layers import (
    align_alternate_layers,
)


DATA = Path(__file__).parent.parent.parent / "data"


@pytest.mark.parametrize(
    "test_path",
    [
        DATA / "AlignAlternateLayers-g2.glyphs",
        DATA / "AlignAlternateLayers-g3.glyphs",
    ],
)
def test_align_alternate_layers(test_path):
    font = GSFont(test_path)
    # 'Cacute' has 3 master layers and no alternate layers
    Cacute = font.glyphs["Cacute"]
    assert len(Cacute.layers) == 3
    assert all(l._is_master_layer for l in Cacute.layers)
    assert not any(l._is_bracket_layer() for l in Cacute.layers)

    # it uses a component 'C' which in turn contains 3 additional alternate layers,
    # plus 'acutecomb.case' which has none
    C = font.glyphs["C"]
    assert len([l for l in font.glyphs["C"].layers if l._is_bracket_layer()]) == 3
    assert not any(l._is_bracket_layer() for l in font.glyphs["acutecomb.case"].layers)

    align_alternate_layers(font)

    # we expect 'Cacute' to now have 3 new alternate layers which have the same
    # axis coordinates as the ones from 'C'
    assert len(Cacute.layers) == 6
    assert len([l for l in Cacute.layers if l._is_master_layer]) == 3
    assert len([l for l in Cacute.layers if l._is_bracket_layer()]) == 3
    assert {
        tuple(l._bracket_axis_rules()) for l in C.layers if l._is_bracket_layer()
    } == {
        tuple(l._bracket_axis_rules()) for l in Cacute.layers if l._is_bracket_layer()
    }



import glyphsLib
from glyphsLib.filters.cornerComponents import CornerComponentsFilter
import py
import pytest


datadir = py.path.local(py.path.local(__file__).dirname).join("data")

ufo = glyphsLib.load_to_ufos(datadir.join("CornerComponents.glyphs"))[0]

test_glyphs = [glyph[:-12] for glyph in ufo.keys() if glyph.endswith(".expectation")]


@pytest.mark.parametrize("glyph", sorted(test_glyphs))
def test_corner_components(glyph):
    if "left_anchor" in glyph:
        pytest.xfail("left anchors not quite working yet")
    philter = CornerComponentsFilter(include={glyph})
    assert philter(ufo)
    test_glyph = ufo[glyph]
    expectation = ufo[glyph + ".expectation"]
    assert len(test_glyph) == len(expectation)
    for test_contour, expectation_contour in zip(
        expectation.contours, test_glyph.contours
    ):
        assert test_contour == expectation_contour, glyph



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
from collections import OrderedDict
from copy import deepcopy

from .constants import BRACKET_GLYPH_RE, UFO_KERN_GROUP_PATTERN


def flip_class_side(s):
    if s.startswith("@MMK_L_"):
        return f"@MMK_R_{s[7:]}"
    elif s.startswith("@MMK_R_"):
        return f"@MMK_L_{s[7:]}"
    return s


def to_ufo_kerning(self):
    for master in self.font.masters:
        kerning_source = master.metricsSource  # Maybe be a linked master
        if kerning_source is None:
            kerning_source = master
        both_directions = (
            kerning_source.id in self.font.kerningLTR
            and kerning_source.id in self.font.kerningRTL
        )
        combined_kerning = OrderedDict()
        if kerning_source.id in self.font.kerningLTR:
            kerning = self.font.kerningLTR[kerning_source.id]
            combined_kerning = deepcopy(kerning) if both_directions else kerning
        if kerning_source.id in self.font.kerningRTL:
            for kern1, subtable in self.font.kerningRTL[kerning_source.id].items():
                # flip RTL sides and combine with existing LTR dicts, but take care
                # not to overwrite whole kern2 subtable when the flipped kern1
                # coincides with an existing LTR kern1
                # https://github.com/googlefonts/glyphsLib/issues/1039
                kern1_key = flip_class_side(kern1)
                existing_kern2 = combined_kerning.setdefault(kern1_key, {})
                new_kern2 = {flip_class_side(kern2): v for kern2, v in subtable.items()}
                # TODO: use 3.9+ dict.update() or | operator after we drop python3.8
                combined_kerning[kern1_key] = {**existing_kern2, **new_kern2}
        if combined_kerning:
            _to_ufo_kerning(self, self._sources[master.id].font, combined_kerning)


def _to_ufo_kerning(self, ufo, kerning_data):
    """Add .glyphs kerning to an UFO."""

    warning_msg = "Non-existent glyph class %s found in kerning rules."

    for left, pairs in kerning_data.items():
        match = re.match(r"@MMK_L_(.+)", left)
        left_is_class = bool(match)
        if left_is_class:
            left = "public.kern1.%s" % match.group(1)
            if left not in ufo.groups:
                self.logger.warning(warning_msg % left)
        for right, kerning_val in pairs.items():
            match = re.match(r"@MMK_R_(.+)", right)
            right_is_class = bool(match)
            if right_is_class:
                right = "public.kern2.%s" % match.group(1)
                if right not in ufo.groups:
                    self.logger.warning(warning_msg % right)
            ufo.kerning[left, right] = kerning_val


def to_glyphs_kerning(self):
    """Add UFO kerning to GSFont."""
    for master_id, source in self._sources.items():
        for (left, right), value in source.font.kerning.items():
            if BRACKET_GLYPH_RE.match(left) or BRACKET_GLYPH_RE.match(right):
                # Skip all bracket glyph entries, as they are duplicates of their
                # parents'.
                continue
            left_match = UFO_KERN_GROUP_PATTERN.match(left)
            right_match = UFO_KERN_GROUP_PATTERN.match(right)
            if left_match:
                left = "@MMK_L_{}".format(left_match.group(2))
            if right_match:
                right = "@MMK_R_{}".format(right_match.group(2))
            self.font.setKerningForPair(master_id, left, right, value)



import pytest
from fontTools.pens.basePen import MissingComponentError

import glyphsLib
from glyphsLib import to_designspace
from glyphsLib.classes import GSComponent


def test_background_component_decompose(datadir):
    font = glyphsLib.GSFont(str(datadir.join("Recursion.glyphs")))
    ds = to_designspace(font, minimal=False)

    for source in ds.sources:
        for layer in source.font.layers:
            for glyph in layer:
                if layer.name == "Apr 27 20, 17:59" and glyph.name == "B":
                    continue
                assert not glyph.components

    ufo_rg = ds.sources[0].font
    assert ufo_rg.layers["public.background"]["A"].contours == ufo_rg["B"].contours
    assert (
        ufo_rg.layers["Apr 27 20, 17:57.background"]["A"].contours
        == ufo_rg["B"].contours
    )
    assert ufo_rg.layers["public.background"]["B"].contours == ufo_rg["A"].contours
    assert len(ufo_rg.layers["Apr 27 20, 17:59.background"]["B"].contours) == 2

    assert ufo_rg.layers["Apr 27 20, 17:59"]["B"].components

    ufo_bd = ds.sources[1].font
    assert ufo_bd.layers["public.background"]["A"].contours == ufo_bd["B"].contours
    assert (
        ufo_bd.layers["Apr 27 20, 17:57.background"]["A"].contours
        == ufo_bd["B"].contours
    )
    assert ufo_bd.layers["public.background"]["B"].contours == ufo_bd["A"].contours
    assert (
        ufo_bd.layers["Apr 27 20, 17:59.background"]["B"].contours
        == ufo_bd["A"].contours
    )


def test_background_component_decompose_missing(datadir):
    font = glyphsLib.GSFont(str(datadir.join("Recursion.glyphs")))

    layer = font.glyphs["B"].layers["DB4D7D04-C02D-48DE-811E-03AA03052DD2"].background
    layer.components.append(GSComponent("xxx"))

    with pytest.raises(MissingComponentError):
        to_designspace(font, minimal=False)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import binascii
import copy
import datetime
import re
from typing import List
from typing import Optional
from typing import Union

__all__ = [
    "BinaryData",
    "IndexPath",
    "Point",
    "Rect",
    "Size",
    "Transform",
    "UnicodesList",
    "ValueType",
    "floatToString3",
    "floatToString5",
    "parse_color",
    "parse_datetime",
    "parse_float_or_int",
    "readIntlist",
]


def parse_float_or_int(value_string):
    v = float(value_string)
    if v.is_integer():
        return int(v)
    return v


class ValueType:
    """A base class for value types that are comparable in the Python sense
    and readable/writable using the glyphsLib parser/writer.
    """

    default = None

    def __init__(self, value=None):
        if value:
            self.value = self.fromString(value)
        else:
            self.value = copy.deepcopy(self.default)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, self.plistValue())

    def fromString(self, src):
        """Return a typed value representing the structured glyphs strings."""
        raise NotImplementedError("%s read" % type(self).__name__)

    def plistValue(self, format_version=2):
        """Return structured glyphs strings representing the typed value."""
        raise NotImplementedError("%s write" % type(self).__name__)

    # https://stackoverflow.com/questions/390250/
    # elegant-ways-to-support-equivalence-equality-in-python-classes
    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(self, other.__class__):
            return self.value == other.value
        return NotImplemented

    def __hash__(self):
        """Overrides the default implementation"""
        return hash(self.value)


# class Vector<dimension>
def Vector(dim):
    class Vector(ValueType):
        """Base type for number vectors (points, rects, transform matrices)."""

        dimension = dim
        default = [0.0] * dimension
        regex = re.compile("[({]%s[})]" % ", ".join(["([-.e\\d]+)"] * dimension))

        def fromString(self, src):
            if isinstance(src, list):
                assert len(src) == self.dimension
                return src
            src = src.replace('"', "")
            return [parse_float_or_int(i) for i in self.regex.match(src).groups()]

        def plistValue(self, format_version=2):
            assert isinstance(self.value, list) and len(self.value) == self.dimension
            if format_version == 2:
                return '"{%s}"' % (", ".join(floatToString3(v) for v in self.value))
            else:
                return "(%s)" % (",".join(floatToString3(v) for v in self.value))

        def __getitem__(self, key):
            assert isinstance(self.value, list) and len(self.value) == self.dimension
            return self.value[key]

        def __setitem__(self, key, value):
            assert isinstance(self.value, list) and len(self.value) == self.dimension
            self.value[key] = value

        def __len__(self):
            return self.dimension

    return Vector


class Point(Vector(2)):
    """Read/write a vector in curly braces."""

    __slots__ = ("value", "rect")

    def __init__(self, value=None, value2=None, rect=None):
        self.rect = rect

        # Invoked like Point(100, 200) or Point(value=100, value2=200).
        if value is not None and value2 is not None:
            self.value = [value, value2]
        # Invoked like Point("{800, 10}").
        elif isinstance(value, str):
            self.value = self.fromString(value)
        # Invoked like Point([100, 200]).
        elif isinstance(value, (list, tuple)):
            self.value = value
        # 🤷
        else:
            raise TypeError(
                "Point must be constructed with two ints or a string representing "
                "a point or a list representing a point."
            )

    def __repr__(self):
        return "<point x={} y={}>".format(self.value[0], self.value[1])

    @property
    def x(self):
        return self.value[0]

    @x.setter
    def x(self, value):
        self.value[0] = value
        # Update parent rect
        if self.rect:
            self.rect.value[0] = value

    @property
    def y(self):
        return self.value[1]

    @y.setter
    def y(self, value):
        self.value[1] = value
        # Update parent rect
        if self.rect:
            self.rect.value[1] = value


class Size(Point):
    def __repr__(self):
        return "<size width={} height={}>".format(self.value[0], self.value[1])

    @property
    def width(self):
        return self.value[0]

    @width.setter
    def width(self, value):
        self.value[0] = value
        # Update parent rect
        if self.rect:
            self.rect.value[2] = value

    @property
    def height(self):
        return self.value[1]

    @height.setter
    def height(self, value):
        self.value[1] = value
        # Update parent rect
        if self.rect:
            self.rect.value[3] = value


class Rect(Vector(4)):
    """Read/write a rect of two points in curly braces."""

    regex = re.compile(r"{{([-.e\d]+), ([-.e\d]+)}, {([-.e\d]+), ([-.e\d]+)}}")

    def __init__(self, value=None, value2=None):
        if value is not None and value2 is not None:
            value = [value[0], value[1], value2[0], value2[1]]
        super().__init__(value)

    def plistValue(self, format_version=2):
        assert isinstance(self.value, list) and len(self.value) == self.dimension
        if format_version == 2:
            return '"{{%s, %s}, {%s, %s}}"' % tuple(
                floatToString3(v) for v in self.value
            )
        return "(" + ",".join(map(floatToString3, self.value)) + ")"

    def __repr__(self):
        return "<rect origin={} size={}>".format(str(self.origin), str(self.size))

    @property
    def origin(self):
        return Point(self.value[0], self.value[1], rect=self)

    @origin.setter
    def origin(self, value):
        self.value[0] = value.x
        self.value[1] = value.y

    @property
    def size(self):
        return Size(self.value[2], self.value[3], rect=self)

    @size.setter
    def size(self, value):
        self.value[2] = value.width
        self.value[3] = value.height


class Transform(Vector(6)):
    """Read/write a six-element vector."""

    def __init__(
        self,
        value=None,
        value2=None,
        value3=None,
        value4=None,
        value5=None,
        value6=None,
    ):
        if all(v is not None for v in (value, value2, value3, value4, value5, value6)):
            value = [value, value2, value3, value4, value5, value6]
        super().__init__(value)

    def __repr__(self):
        return "<affine transformation %s>" % (" ".join(map(str, self.value)))

    def plistValue(self, format_version=2):
        assert isinstance(self.value, list) and len(self.value) == self.dimension
        return '"{%s}"' % (", ".join(floatToString5(v) for v in self.value))

    def determinant(self):
        a, b, c, d = self[:4]
        return a * d - b * c


UTC_OFFSET_RE = re.compile(r".* (?P<sign>[+-])(?P<hours>\d\d)(?P<minutes>\d\d)$")


def parse_datetime(src=None):
    """Parse a datetime object from a string."""
    if src is None:
        return None
    string = src.replace('"', "")
    # parse timezone ourselves, since %z is not always supported
    # see: http://bugs.python.org/issue6641
    m = UTC_OFFSET_RE.match(string)
    if m:
        sign = 1 if m.group("sign") == "+" else -1
        tz_hours = sign * int(m.group("hours"))
        tz_minutes = sign * int(m.group("minutes"))
        offset = datetime.timedelta(hours=tz_hours, minutes=tz_minutes)
        string = string[:-6]
    else:
        # no explicit timezone
        offset = datetime.timedelta(0)
    if "AM" in string or "PM" in string:
        datetime_obj = datetime.datetime.strptime(string, "%Y-%m-%d %I:%M:%S %p")
    else:
        datetime_obj = datetime.datetime.strptime(string, "%Y-%m-%d %H:%M:%S")
    return datetime_obj + offset


# FIXME: (jany) Not sure this should be used
class Datetime(ValueType):
    """Read/write a datetime.  Doesn't maintain time zone offset."""

    def fromString(self, src):
        return parse_datetime(src)

    def plistValue(self, format_version=2):
        return '"%s +0000"' % self.value

    def strftime(self, val):
        return self.value.strftime(val)


def parse_color(src=None):
    """Parse a string representing a color value.

    Color is either a fixed color (when coloring something from the UI, see
    the GLYPHS_COLORS constant) or a list of the format [u8, u8, u8, u8],

    Glyphs does not support an alpha channel as of 2.5.1 (confirmed by Georg
    Seifert), and always writes a 1 to it. This was brought up and is probably
    corrected in the next versions.
    https://github.com/googlefonts/glyphsLib/pull/363#issuecomment-390418497
    """
    if src is None:
        return None

    # Tuple.
    if src[0] == "(":
        rgba = tuple(int(v) for v in src[1:-1].split(",") if v)

        if not (len(rgba) == 4 and all(0 <= v < 256 for v in rgba)):
            raise ValueError(
                "Broken color tuple: {}. Must have four values from 0 to 255.".format(
                    src
                )
            )

        return rgba

    # Constant.
    return int(src)


# FIXME: (jany) not sure this is used
class Color(ValueType):
    def fromString(self, src):
        return parse_color(src)

    def __repr__(self):
        return self.value.__repr__()

    def plistValue(self, format_version=2):
        return str(self.value)


# mutate list in place
def _mutate_list(fn, l):
    assert isinstance(l, list)
    for i in range(len(l)):
        l[i] = fn(l[i])
    return l


def readIntlist(src):
    return _mutate_list(int, src)


def writeIntlist(val):
    return _mutate_list(str, val)


def floatToString3(f: float) -> str:
    """Return float f as a string with three decimal places without trailing zeros
    and dot.

    Intended for places where three decimals are enough, e.g. node positions.
    """
    return f"{f:.3f}".rstrip("0").rstrip(".")


def floatToString5(f: float) -> str:
    """Return float f as a string with five decimal places without trailing zeros
    and dot.

    Intended for places where five decimals are needed, e.g. transformations.
    """
    return f"{f:.5f}".rstrip("0").rstrip(".")


class UnicodesList(list):
    """Represent a PLIST-able list of unicode codepoints as strings."""

    def __init__(self, value=None):
        if value is None:
            unicodes = []
        elif isinstance(value, str):
            unicodes = value.split(",")
        else:
            unicodes = [str(v) for v in value]
        super().__init__(unicodes)

    def plistValue(self, format_version=2):
        if not self:
            return None
        if len(self) == 1:
            if format_version == 3:
                return str(int(self[0], 16))
            return self[0]
        if format_version == 2:
            return '"%s"' % ",".join(self)
        else:
            return "(%s)" % ",".join([str(int(x, 16)) for x in self])


class BinaryData(bytes):
    @classmethod
    def fromHex(cls, data):
        return cls(binascii.unhexlify(data))

    def plistValue(self, format_version=2):
        # TODO write hex bytes in chunks and split over multiple lines
        # for better readability, like the fonttools xmlWriter does
        return "<%s>" % binascii.hexlify(self).decode()


class IndexPath(ValueType):
    """A list of indexes.

    It is analogous to `NSIndexPath`, which is a list of indices that together
    represent the path to a specific location in a tree of nested arrays. This
    class is used internally by Glyphs for storing properties of hints,
    including `origin`, `other1`, `other2`, and `target`.

    The most common case is that it is a list of two integers pointing at a real
    node. However, it can also be three or four integers. Moreover, in the case
    of `origin` and `target`, it can be a list containing a single string.
    """

    def __init__(
        self,
        value: Union[int, str, List[Union[int, str]]],
        value2: Optional[int] = None,
        value3: Optional[int] = None,
        value4: Optional[int] = None,
    ):
        if value4 is not None:
            self.value = [value, value2, value3, value4]
        elif value3 is not None:
            self.value = [value, value2, value3]
        elif value2 is not None:
            self.value = [value, value2]
        elif isinstance(value, (list, tuple)):
            self.value = value
        else:
            self.value = self.fromString(value)

    def fromString(self, string: str) -> List[Union[int, str]]:
        values = string.strip().lstrip('"{(').rstrip(')}"').split(",")
        return [
            (int(value) if value.isdigit() else value)
            for value in map(str.strip, values)
        ]

    def plistValue(self, format_version: int = 2) -> str:
        if len(self.value) == 1:
            return self.value[0]
        if format_version == 2:
            return '"{' + ", ".join(map(str, self.value)) + '}"'
        return "(" + ",".join(map(str, self.value)) + ")"

    def __getitem__(self, key: int) -> Union[int, str]:
        return self.value[key]

    def __setitem__(self, key: int, value: Union[int, str]) -> None:
        self.value[key] = value

    def __len__(self) -> int:
        return len(self.value)



"""This module is DEPRECATED and will be removed in a future release.

For anchor propagation on GSFont objects, you can use the
`glyphsLib.builder.transformations.propagate_anchors` module.
For anchor propagation on UFO font objects, you can try the
`ufo2ft.filters.propagateAnchors` filter.
"""

from fontTools.misc.transform import Transform
import fontTools.pens.boundsPen

from .constants import COMPONENT_INFO_KEY


def to_ufo_propagate_font_anchors(self, ufo):
    """Copy anchors from parent glyphs' components to the parent."""

    processed = set()
    for glyph in ufo:
        _propagate_glyph_anchors(self, ufo, glyph, processed)


def _propagate_glyph_anchors(self, ufo, parent, processed):
    """Propagate anchors for a single parent glyph."""

    if parent.name in processed:
        return
    processed.add(parent.name)

    base_components = []
    mark_components = []
    anchor_names = set()
    to_add = {}
    for component in parent.components:
        try:
            glyph = ufo[component.baseGlyph]
        except KeyError:
            self.logger.warning(
                "Anchors not propagated for inexistent component {} in glyph {}".format(
                    component.baseGlyph, parent.name
                )
            )
        else:
            _propagate_glyph_anchors(self, ufo, glyph, processed)
            if any(a.name.startswith("_") for a in glyph.anchors):
                mark_components.append(component)
            else:
                base_components.append(component)
                anchor_names |= {a.name for a in glyph.anchors}

    if mark_components and not base_components and _is_ligature_mark(parent):
        # The composite is a mark that is composed of other marks (E.g.
        # "circumflexcomb_tildecomb"). Promote the mark that is positioned closest
        # to the origin to a base.
        try:
            component = _component_closest_to_origin(mark_components, ufo)
        except Exception as e:
            raise Exception(
                "Error while determining which component of composite "
                "'{}' is the lowest: {}".format(parent.name, str(e))
            ) from e
        mark_components.remove(component)
        base_components.append(component)
        glyph = ufo[component.baseGlyph]
        anchor_names |= {a.name for a in glyph.anchors}

    for anchor_name in anchor_names:
        # don't add if parent already contains this anchor OR any associated
        # ligature anchors (e.g. "top_1, top_2" for "top")
        if not any(a.name.startswith(anchor_name) for a in parent.anchors):
            _get_anchor_data(to_add, ufo, base_components, anchor_name)

    for component in mark_components:
        _adjust_anchors(to_add, ufo, parent, component)

    # we sort propagated anchors to append in a deterministic order
    for name, (x, y) in sorted(to_add.items()):
        anchor_dict = {"name": name, "x": x, "y": y}
        parent.appendAnchor(anchor_dict)


def _get_anchor_data(anchor_data, ufo, components, anchor_name):
    """Get data for an anchor from a list of components."""

    anchors = []
    for component in components:
        for anchor in ufo[component.baseGlyph].anchors:
            if anchor.name == anchor_name:
                anchors.append((anchor, component))
                break
    if len(anchors) > 1:
        for i, (anchor, component) in enumerate(anchors):
            t = Transform(*component.transformation)
            name = "%s_%d" % (anchor.name, i + 1)
            anchor_data[name] = t.transformPoint((anchor.x, anchor.y))
    elif anchors:
        anchor, component = anchors[0]
        t = Transform(*component.transformation)
        anchor_data[anchor.name] = t.transformPoint((anchor.x, anchor.y))


def _componentAnchorFromLib(_glyph, _targetComponent):
    """Pull component’s named anchor from Glyph.lib"""
    if COMPONENT_INFO_KEY in _glyph.lib:
        for _anchorLib in _glyph.lib[COMPONENT_INFO_KEY]:
            if (
                "anchor" in _anchorLib
                and "name" in _anchorLib
                and _anchorLib["name"] == _targetComponent.baseGlyph
                and _anchorLib["index"] == _glyph.components.index(_targetComponent)
            ):
                return _anchorLib["anchor"] or None
    return None


def _adjust_anchors(anchor_data, ufo, parent, component):
    """Adjust anchors to which a mark component may have been attached."""
    glyph = ufo[component.baseGlyph]
    anchor_names = {a.name for a in glyph.anchors}
    t = Transform(*component.transformation)
    component_anchor = _componentAnchorFromLib(parent, component)
    # ignore the component's named anchor if we don't have it
    if component_anchor not in anchor_data:
        component_anchor = None
    # For each base anchor in the mark component glyph
    for anchor in (a for a in glyph.anchors if not a.name.startswith("_")):
        # adjust either if component is attached to a specific named anchor
        # (e.g. top_2 for a ligature glyph)
        # rather than to the standard anchors (top/bottom)
        if (
            component_anchor
            and component_anchor.startswith(anchor.name + "_")
            and "_" + anchor.name in anchor_names
        ):
            anchor_data[component_anchor] = t.transformPoint((anchor.x, anchor.y))
        # ... or this anchor has data and the component also contains
        # the associated mark anchor (e.g. "_top" for "top") ...
        elif anchor.name in anchor_data and "_" + anchor.name in anchor_names:
            anchor_data[anchor.name] = t.transformPoint((anchor.x, anchor.y))


def _is_ligature_mark(glyph):
    return not glyph.name.startswith("_") and "_" in glyph.name


def _component_closest_to_origin(components, glyph_set):
    """Return the component whose (xmin, ymin) bounds are closest to origin.

    This ensures that a component that is moved below another is
    actually recognized as such. Looking only at the transformation
    offset can be misleading.
    """
    return min(components, key=lambda comp: _distance((0, 0), _bounds(comp, glyph_set)))


def _distance(pos1, pos2):
    x1, y1 = pos1
    x2, y2 = pos2
    return (x1 - x2) ** 2 + (y1 - y2) ** 2


def _bounds(component, glyph_set):
    """Return the (xmin, ymin) of the bounds of `component`."""
    if hasattr(component, "bounds"):  # e.g. defcon
        return component.bounds[:2]
    elif hasattr(component, "draw"):  # e.g. ufoLib2
        pen = fontTools.pens.boundsPen.BoundsPen(glyphSet=glyph_set)
        component.draw(pen)
        return pen.bounds[:2]
    else:
        raise ValueError(
            "Don't know to to compute the bounds of component '{}' ".format(component)
        )



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
import logging

import glyphsLib.glyphdata

from .. import GSLayer, GSPath, GSComponent
from .common import from_loose_ufo_time, to_ufo_time
from .constants import (
    GLYPHLIB_PREFIX,
    GLYPHS_COLORS,
    PUBLIC_PREFIX,
    UFO2FT_COLOR_LAYER_MAPPING_KEY,
    BRACKET_GLYPH_RE,
    BRACKET_GLYPH_SUFFIX_RE,
    SCRIPT_LIB_KEY,
    SHAPE_ORDER_LIB_KEY,
    ORIGINAL_WIDTH_KEY,
    BACKGROUND_WIDTH_KEY,
)

logger = logging.getLogger(__name__)


def _clone_layer(layer, paths=None, components=None):
    paths = paths if paths is not None else []
    components = components if components is not None else []
    if len(paths) == len(layer.paths) and len(components) == len(layer.components):
        return layer
    new_layer = GSLayer()
    new_layer.associatedMasterId = layer.associatedMasterId
    new_layer.parent = layer.parent
    new_layer.paths = paths
    new_layer.components = components
    new_layer.attributes = layer.attributes
    return new_layer


# Map of ".uvNNN" extensions to Unicode Variation Selector code points.
# If a glyph name ends with ".uvNNN" with NNN ranging from 001 to 256, then it
# is a variation sequence with uv001 being U+0xFE00 and uv256 being U+0xE01EF.
#
# The only documentation for this is the "More Improvements" section in Glyphs
# 2.6.1 announcement:
# https://glyphsapp.com/news/glyphs-2-6-1-released
# And this forum post:
# https://forum.glyphsapp.com/t/unicode-variation-selector-u-fe01/21701
USV_MAP = {
    f".uv{i + 1:03}": f"{c:04X}"
    for i, c in enumerate(
        itertools.chain(range(0xFE00, 0xFE0F + 1), range(0xE0100, 0xE01EF + 1))
    )
}

USV_EXTENSIONS = tuple(USV_MAP.keys())


def to_ufo_glyph(  # noqa: C901
    self, ufo_glyph, layer, glyph, do_color_layers=True, is_color_layer_glyph=False
):
    """Add .glyphs metadata, paths, components, and anchors to a glyph."""
    ufo_font = self._sources[layer.associatedMasterId or layer.layerId].font

    if layer.layerId == layer.associatedMasterId and do_color_layers:
        self.to_ufo_glyph_color(ufo_glyph, layer, glyph)

    # Color layer glyphs (e.g., "Acaron.color0") should not get unicode values
    # from their parent glyph to avoid duplicate unicode mappings.
    if not is_color_layer_glyph:
        ufo_glyph.unicodes = [int(uval, 16) for uval in glyph.unicodes]

    export = glyph.export
    if not export:
        if self.write_skipexportglyphs:
            # Note that we use `ufo_glyph.name`, not `glyph.name`; these are normally
            # the same however for bracket layers, the name of the UFO glyph has a
            # .varAltNN suffix.
            self.skip_export_glyphs.add(ufo_glyph.name)
        else:
            ufo_glyph.lib[GLYPHLIB_PREFIX + "Export"] = export

    # If glyph name ends with ".uvNNN" find and the font has a glyph with the
    # same name without the ".uvNNN", then add a Unicode Variation Sequence
    # entry with the Unicode Variation Selector corresponding to the extension
    # and the unicode of the base glyph.
    if export and "." in glyph.name and glyph.name.endswith(USV_EXTENSIONS):
        base_name, ext = glyph.name.rsplit(".", 1)
        if base_name in glyph.parent.glyphs and glyph.parent.glyphs[base_name].unicode:
            uni = glyph.parent.glyphs[base_name].unicode
            usv = USV_MAP[f".{ext}"]
            USV_KEY = PUBLIC_PREFIX + "unicodeVariationSequences"
            ufo_font.lib.setdefault(USV_KEY, {}).setdefault(usv, {})[uni] = glyph.name

    # we can't use use the glyphs.unicodes values since they aren't always
    # correctly padded
    unicodes = [f"{c:04X}" for c in ufo_glyph.unicodes]
    # FIXME: (jany) next line should be an API of GSGlyph?
    glyphinfo = glyphsLib.glyphdata.get_glyph(ufo_glyph.name, unicodes=unicodes)

    if self.glyphdata is not None:
        custom = glyphsLib.glyphdata.get_glyph(
            ufo_glyph.name, self.glyphdata, unicodes=unicodes
        )
        production_name = glyph.production or (
            custom.production_name
            if custom.production_name != glyphinfo.production_name
            else None
        )
        category = glyph.category or (
            custom.category if custom.category != glyphinfo.category else None
        )
        subCategory = glyph.subCategory or (
            custom.subCategory if custom.subCategory != glyphinfo.subCategory else None
        )
        script = glyph.script or (
            custom.script if custom.script != glyphinfo.script else None
        )
    else:
        production_name, category, subCategory, script = (
            glyph.production,
            glyph.category,
            glyph.subCategory,
            glyph.script,
        )

    production_name = production_name or glyphinfo.production_name

    if production_name:
        # Make sure production names of bracket glyphs also get a BRACKET suffix.
        bracket_glyph_name = BRACKET_GLYPH_RE.match(ufo_glyph.name)
        prod_bracket_glyph_name = BRACKET_GLYPH_RE.match(production_name)
        if bracket_glyph_name and not prod_bracket_glyph_name:
            production_name += BRACKET_GLYPH_SUFFIX_RE.match(ufo_glyph.name).group(1)
    if production_name and production_name != ufo_glyph.name:
        postscriptNamesKey = PUBLIC_PREFIX + "postscriptNames"
        if postscriptNamesKey not in ufo_font.lib:
            ufo_font.lib[postscriptNamesKey] = dict()
        ufo_font.lib[postscriptNamesKey][ufo_glyph.name] = production_name

    if script:
        ufo_glyph.lib[SCRIPT_LIB_KEY] = script

    # if glyph contains custom 'category' and 'subCategory' overrides, store
    # them in the UFO glyph's lib
    if category:
        ufo_glyph.lib[GLYPHLIB_PREFIX + "category"] = category
    else:
        category = glyphinfo.category
    if subCategory:
        ufo_glyph.lib[GLYPHLIB_PREFIX + "subCategory"] = subCategory
    else:
        subCategory = glyphinfo.subCategory

    # load width before background, which is loaded with lib data

    width = effective_width(layer, glyph)
    if width is None:
        pass
    elif category == "Mark" and subCategory == "Nonspacing" and width > 0:
        # zero the width of Nonspacing Marks like Glyphs.app does on export
        # TODO: (jany) check for customParameter DisableAllAutomaticBehaviour
        # FIXME: (jany) also don't do that when rt UFO -> glyphs -> UFO
        ufo_glyph.lib[ORIGINAL_WIDTH_KEY] = width
        ufo_glyph.width = 0
    else:
        ufo_glyph.width = width

    if not self.minimal:
        to_ufo_glyph_roundtripping(ufo_glyph, glyph, layer)  # below
        self.to_ufo_background_image(ufo_glyph, layer)  # .background_image
        self.to_ufo_guidelines(ufo_glyph, layer)  # .guidelines
        self.to_ufo_glyph_background(ufo_glyph, layer)  # below
        self.to_ufo_annotations(ufo_glyph, layer)  # .annotations
        self.to_ufo_smart_component_axes(ufo_glyph, glyph)  # .components
    self.to_ufo_glyph_user_data(ufo_font, ufo_glyph, glyph)  # .user_data
    self.to_ufo_layer_user_data(ufo_glyph, layer)  # .user_data

    # Optimization: profiling glyphs2ufo of NotoSans-MM.glyphs (6000 glyphs) on a Mac
    # mini late 2014, Python 3.6.8, revealed that a whopping 17% of the time was spent
    # converting lastChange to UFO timestamps. I could not reproduce this on a Windows
    # 10/Python 3.7 setup, so this might be a platform thing. If-guarding anyway
    # because these timestamps are useless in a UFO scenario if you use Git.

    # This *should* be gated with not self.minimal, but that breaks regression tests
    if (
        self.minimize_glyphs_diffs
        and glyph.parent.customParameters["Disable Last Change"] is not True
        and glyph.lastChange is not None
    ):
        ufo_glyph.lib[GLYPHLIB_PREFIX + "lastChange"] = to_ufo_time(glyph.lastChange)

    self.to_ufo_hints(ufo_glyph, layer)  # .hints
    self.to_ufo_paths(ufo_glyph, layer)  # .paths
    self.to_ufo_components(ufo_glyph, layer)  # .components
    # Store shape order for mixed glyphs
    if layer.paths and layer.components:
        ufo_glyph.lib[SHAPE_ORDER_LIB_KEY] = ""
        for shape in layer.shapes:
            if isinstance(shape, GSPath):
                ufo_glyph.lib[SHAPE_ORDER_LIB_KEY] += "P"
            elif isinstance(shape, GSComponent):
                ufo_glyph.lib[SHAPE_ORDER_LIB_KEY] += "C"
            else:
                raise ValueError("Unknown shape type %s" % shape)
    self.to_ufo_glyph_anchors(ufo_glyph, layer.anchors)  # .anchors
    if self.is_vertical:
        self.to_ufo_glyph_height_and_vertical_origin(ufo_glyph, layer)  # below


def to_ufo_glyph_roundtripping(ufo_glyph, glyph, layer):
    note = glyph.note
    if note is not None:
        ufo_glyph.note = note

    color_index = glyph.color
    if color_index is not None:
        # .3f is enough precision to round-trip uint8 to float losslessly.
        # https://github.com/unified-font-object/ufo-spec/issues/61
        # #issuecomment-389759127
        if (
            isinstance(color_index, list)
            and len(color_index) == 4
            and all(0 <= v < 256 for v in color_index)
        ):
            ufo_glyph.markColor = ",".join(
                "0" if v == 0 else "1" if v == 255 else "{:.3f}".format(v / 255)
                for v in color_index
            )
        elif isinstance(color_index, int) and color_index in range(len(GLYPHS_COLORS)):
            ufo_glyph.markColor = GLYPHS_COLORS[color_index]
        else:
            logger.warning(
                "Glyph {}, layer {}: Invalid color index/tuple {}".format(
                    glyph.name, layer.name, color_index
                )
            )

    for key in ["leftMetricsKey", "rightMetricsKey", "widthMetricsKey"]:
        value = getattr(layer, key, None)
        if value:
            ufo_glyph.lib[GLYPHLIB_PREFIX + "layer." + key] = value
        value = getattr(glyph, key, None)
        if value:
            ufo_glyph.lib[GLYPHLIB_PREFIX + "glyph." + key] = value


def effective_width(layer, glyph):
    # The width may be taken from another master via the customParameters
    # 'Link Metrics With Master' or 'Link Metrics With First Master'.
    font = glyph.parent
    master = font.masters[layer.layerId]
    if master:
        metrics_source = master.metricsSource
        if metrics_source:
            metric_layer = font.glyphs[glyph.name].layers[metrics_source.id]
            if metric_layer:
                width = metric_layer.width
                if layer.width != width:
                    logger.debug(
                        f"{layer.parent.name}: Applying width from master "
                        f"'{metrics_source.id}': {layer.width} -> {width}"
                    )
                return width
    return layer.width


def to_ufo_glyph_color(self, ufo_glyph, layer, glyph, do_color_layers=True):
    # Here we handle color layers. If this is a master layer and the glyph
    # has color layers, add ufo2ft lib key with the layer mapping.

    # There are two kinds of color layers: first, color palette layers that
    # are handled below, which are used to build COLRv0 table. For color
    # palette layers, the layer mapping is a tuple of (layer name, palette
    # index), but we don’t know the final UFO layer names yet, so we use
    # Glyphs layer IDs and change them to layer names in
    # to_ufo_color_layer_names().
    # When building minimal UFOs, we instead collect color layers and later
    # add them as separate glyphs to the UFO font.

    if any(
        l._is_color_palette_layer() and l.associatedMasterId == layer.associatedMasterId
        for l in glyph.layers
    ):
        layerMapping = [
            (l.layerId, l._color_palette_index())
            for l in glyph.layers
            if l._is_color_palette_layer()
            and l.associatedMasterId == layer.associatedMasterId
        ]

        if not self.minimal:
            ufo_glyph.lib[UFO2FT_COLOR_LAYER_MAPPING_KEY] = layerMapping
        elif glyph.export:
            layers = []
            for layerId, colorId in layerMapping:
                layers.append((glyph.layers[layerId], colorId))
            self._color_palette_layers.append(((glyph, layer), layers))

    if self.minimal:
        # The other kind of color layers supports solid colors and
        # gradients among other things, and we use it to build COLRv1
        # table.
        # For each color layer, we collect paths that has the same
        # attributes, then we make a clone of the layer for each group with
        # only the paths in this group. We do this splitting because a
        # COLRv1 layer can’t have multiple gradients or colors.
        color_layers = [
            l
            for l in glyph.layers
            if l.attributes.get("color")
            and l.associatedMasterId == layer.associatedMasterId
        ]
        if color_layers:
            layers = []
            for color_layer in color_layers:
                # Group consecutive paths with same attributes together.
                groups = [
                    list(g)
                    for k, g in itertools.groupby(
                        color_layer.paths, key=lambda p: p.attributes
                    )
                ]
                for paths in groups:
                    layers.append(_clone_layer(color_layer, paths=paths))

                # Group components based on whether component glyph has
                # color layers or not.
                groups = [
                    (k, list(g))
                    for k, g in itertools.groupby(
                        color_layer.components,
                        key=lambda c: any(
                            l.attributes.get("color")
                            for l in c.component.layers
                            if l.associatedMasterId == layer.associatedMasterId
                        ),
                    )
                ]
                for has_color, components in groups:
                    if not has_color:
                        new_layer = _clone_layer(color_layer, components=components)
                        new_layer.attributes = {}
                        layers.append(new_layer)
                    else:
                        for c in components:
                            layers.append(_clone_layer(color_layer, components=[c]))

            self._color_layers.append(((glyph, layer), layers))


def to_ufo_glyph_height_and_vertical_origin(self, ufo_glyph, layer):
    # implentation based on:
    # https://github.com/googlefonts/glyphsLib/issues/557#issuecomment-667074856
    assert self.is_vertical

    ascender, descender = _get_typo_ascender_descender(layer.master)

    if layer.vertWidth is not None:
        ufo_glyph.height = layer.vertWidth
    else:
        ufo_glyph.height = ascender - descender

    if layer.vertOrigin is not None:
        ufo_glyph.verticalOrigin = ascender - layer.vertOrigin
    else:
        ufo_glyph.verticalOrigin = ascender


def _get_typo_ascender_descender(master):
    # Glyphsapp will use the typo metrics to set the verOrigin and
    # vertWidth. If typo metrics are not present, the master
    # ascender and descender are used instead.
    if "typoAscender" in master.customParameters:
        ascender = master.customParameters["typoAscender"]
    else:
        ascender = master.ascender
    if "typoDescender" in master.customParameters:
        descender = master.customParameters["typoDescender"]
    else:
        descender = master.descender
    return ascender, descender


def to_ufo_glyph_background(self, glyph, layer):
    """Set glyph background."""

    if not layer.hasBackground:
        return

    background = layer.background
    ufo_layer = self.to_ufo_background_layer(layer)
    new_glyph = ufo_layer.newGlyph(glyph.name)

    width = background.userData[BACKGROUND_WIDTH_KEY]
    if width is not None:
        new_glyph.width = width

    self.to_ufo_background_image(new_glyph, background)
    self.to_ufo_paths(new_glyph, background)
    self.to_ufo_components(new_glyph, background)
    self.to_ufo_glyph_anchors(new_glyph, background.anchors)
    self.to_ufo_guidelines(new_glyph, background)


# UFO to Glyphs


def to_glyphs_glyph(self, ufo_glyph, ufo_layer, master):  # noqa: C901
    """Add UFO glif metadata, paths, components, and anchors to a GSGlyph.
    If the matching GSGlyph does not exist, then it is created,
    else it is updated with the new data.
    In all cases, a matching GSLayer is created in the GSGlyph to hold paths.
    """

    # FIXME: (jany) split between glyph and layer attributes
    #        have a write the first time, compare the next times for glyph
    #        always write for the layer

    # NOTE: This optimizes around the performance drain that is glyph name lookup
    #       without replacing the actual data structure. Ideally, FontGlyphsProxy
    #       provides O(1) lookup for all the ways you can use strings to look up
    #       glyphs.
    ufo_glyph_name = ufo_glyph.name  # Avoid method lookup in hot loop.
    glyph = None
    for glyph_object in self.font._glyphs:  # HOT LOOP. Avoid FontGlyphsProxy for speed!
        if glyph_object.name == ufo_glyph_name:  # HOT HOT HOT
            glyph = glyph_object
            break
    if glyph is None:
        glyph = self.glyphs_module.GSGlyph(name=ufo_glyph_name)
        # FIXME: (jany) ordering?
        self.font.glyphs.append(glyph)

    if ufo_glyph.unicodes:
        glyph.unicodes = [f"{c:04X}" for c in ufo_glyph.unicodes]
    glyph.note = ufo_glyph.note or ""
    if GLYPHLIB_PREFIX + "lastChange" in ufo_glyph.lib:
        last_change = ufo_glyph.lib[GLYPHLIB_PREFIX + "lastChange"]
        # We cannot be strict about the dateformat because it's not an official
        # UFO field mentioned in the spec so it could happen to have a timezone
        glyph.lastChange = from_loose_ufo_time(last_change)
    if ufo_glyph.markColor:
        glyph.color = _to_glyphs_color(ufo_glyph.markColor)

    # The export flag can be stored in the glyph's lib key (for upgrading legacy
    # sources) or the Designspace-level public.skipExportGlyphs lib key (canonical
    # place to store the information). The UFO level lib key is ignored.
    if GLYPHLIB_PREFIX + "Export" in ufo_glyph.lib:
        glyph.export = ufo_glyph.lib[GLYPHLIB_PREFIX + "Export"]
    if ufo_glyph.name in self.skip_export_glyphs:
        glyph.export = False

    ufo_font = self._sources[master.id].font
    ps_names_key = PUBLIC_PREFIX + "postscriptNames"
    if ps_names_key in ufo_font.lib and ufo_glyph.name in ufo_font.lib[ps_names_key]:
        glyph.production = ufo_font.lib[ps_names_key][ufo_glyph.name]
        # FIXME: (jany) maybe put something in glyphinfo? No, it's readonly
        #        maybe don't write in glyph.production if glyphinfo already
        #        has something
        # glyphinfo = glyphsLib.glyphdata.get_glyph(ufo_glyph.name)
        # production_name = glyph.production or glyphinfo.production_name

    glyphinfo = glyphsLib.glyphdata.get_glyph(ufo_glyph.name)

    layer = self.to_glyphs_layer(ufo_layer, glyph, master)

    for key in ["leftMetricsKey", "rightMetricsKey", "widthMetricsKey"]:
        # Also read the old version of the key that didn't have a prefix and
        # store it on the layer (because without the "glyph"/"layer" prefix we
        # didn't know whether it originally came from the layer of the glyph,
        # so it's easier to put it back on the most specific level, i.e. the
        # layer)
        for prefix, glyphs_object in (
            ("glyph.", glyph),
            ("", layer),
            ("layer.", layer),
        ):
            full_key = GLYPHLIB_PREFIX + prefix + key
            if full_key in ufo_glyph.lib:
                value = ufo_glyph.lib[full_key]
                setattr(glyphs_object, key, value)

    if SCRIPT_LIB_KEY in ufo_glyph.lib:
        glyph.script = ufo_glyph.lib[SCRIPT_LIB_KEY]

    if GLYPHLIB_PREFIX + "category" in ufo_glyph.lib:
        # TODO: (jany) store category only if different from glyphinfo?
        category = ufo_glyph.lib[GLYPHLIB_PREFIX + "category"]
        glyph.category = category
    else:
        category = glyphinfo.category
    if GLYPHLIB_PREFIX + "subCategory" in ufo_glyph.lib:
        sub_category = ufo_glyph.lib[GLYPHLIB_PREFIX + "subCategory"]
        glyph.subCategory = sub_category
    else:
        sub_category = glyphinfo.subCategory

    # load width before background, which is loaded with lib data
    if hasattr(layer, "foreground"):
        if ufo_glyph.width:
            # Don't store "0", it's the default in UFO.
            # Store in userData because the background's width is not relevant
            # in Glyphs.
            layer.userData[BACKGROUND_WIDTH_KEY] = ufo_glyph.width
    else:
        layer.width = ufo_glyph.width
    if category == "Mark" and sub_category == "Nonspacing" and layer.width == 0:
        # Restore originalWidth
        if ORIGINAL_WIDTH_KEY in ufo_glyph.lib:
            layer.width = ufo_glyph.lib[ORIGINAL_WIDTH_KEY]
            # TODO: (jany) check for customParam DisableAllAutomaticBehaviour?

    self.to_glyphs_background_image(ufo_glyph, layer)
    self.to_glyphs_guidelines(ufo_glyph, layer)
    self.to_glyphs_annotations(ufo_glyph, layer)
    self.to_glyphs_hints(ufo_glyph, layer)
    self.to_glyphs_glyph_user_data(ufo_font, glyph)
    self.to_glyphs_layer_user_data(ufo_glyph, layer)
    self.to_glyphs_smart_component_axes(ufo_glyph, glyph)

    self.to_glyphs_paths(ufo_glyph, layer)
    self.to_glyphs_components(ufo_glyph, layer)

    if SHAPE_ORDER_LIB_KEY in ufo_glyph.lib:
        # Reshuffle shapes array to match original shape order
        new_shapes = []
        path_counter = 0
        comp_counter = 0
        for sign in ufo_glyph.lib[SHAPE_ORDER_LIB_KEY]:
            if sign == "P":
                new_shapes.append(layer.paths[path_counter])
                path_counter += 1
            elif sign == "C":
                new_shapes.append(layer.components[comp_counter])
                comp_counter += 1
            else:
                raise ValueError("Unknown shape type %s" % sign)
        layer.shapes = new_shapes

    self.to_glyphs_glyph_anchors(ufo_glyph, layer)
    self.to_glyphs_glyph_height_and_vertical_origin(ufo_glyph, master, layer)


def _to_glyphs_color(color):
    # If the color matches one of Glyphs's predefined colors, return that
    # index.
    for index, glyphs_color in enumerate(GLYPHS_COLORS):
        if str(color) == glyphs_color:
            return index

    # Otherwise, make a Glyphs-formatted RGBA color list: [u8, u8, u8, u8].
    # Glyphs up to version 2.5.1 always set the alpha channel to 1. It should
    # round-trip the actual value in later versions.
    # https://github.com/googlefonts/glyphsLib/pull/363#issuecomment-390418497
    return [round(float(component) * 255) for component in color.split(",")]


def to_glyphs_glyph_height_and_vertical_origin(self, ufo_glyph, master, layer):
    ascender, descender = _get_typo_ascender_descender(master)
    if ufo_glyph.height != (ascender - descender):
        layer.vertWidth = ufo_glyph.height

    if ufo_glyph.verticalOrigin is not None and ufo_glyph.verticalOrigin != ascender:
        layer.vertOrigin = ascender - ufo_glyph.verticalOrigin



# Copyright 2021 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from fontTools.misc.transform import Identity, Transform
from fontTools.ttLib.tables.otTables import PaintFormat

from .common import to_ufo_color
from .constants import UFO2FT_COLOR_LAYERS_KEY, UFO2FT_COLOR_PALETTES_KEY


def _to_ufo_color_palette_layers(builder, master, layerMapping):
    for (glyph, masterLayer), layers in builder._color_palette_layers:
        if master.id != masterLayer.associatedMasterId:
            continue

        colorLayers = []
        for i, (layer, colorId) in enumerate(layers):
            if layer.layerId == master.id:
                # This is master layer, we can re-use its UFO glyph.
                layerGlyphName = glyph.name
            else:
                # Not the master layer, create a new UFO glyph for it.
                layerGlyphName = f"{glyph.name}.color{i}"
                ufo_layer = builder.to_ufo_layer(glyph, masterLayer)
                ufo_glyph = ufo_layer.newGlyph(layerGlyphName)
                builder.to_ufo_glyph(ufo_glyph, layer, glyph, is_color_layer_glyph=True)
            colorLayers.append((layerGlyphName, colorId))
        layerMapping[glyph.name] = colorLayers


def _find_or_insert_color(color, palette):
    if color is None:
        return 0xFFFF, 1
    color = to_ufo_color(color)
    palette, old = palette
    if color not in palette:
        palette.append(color)
    return len(old) + palette.index(color), color[-1]


def _radius(rect, point):
    # Emulate how AppKit’s “drawInRect:relativeCenterPosition:” calculates the
    # radius, since this is what Glyphs uses.
    center = (rect.width * point[0], rect.height * point[1])
    distances = []
    for pt in ((0, 0), (rect.width, 0), (0, rect.height), (rect.width, rect.height)):
        dist = math.dist(center, pt)
        distances.append(dist)
    return max(distances)


def _to_gradient_paint(gradient, layer, palette):
    # Glyphs start and stop points seem to be a percentage of the
    # path size, but we want absolute coordinates.
    bounds = layer.bounds
    x0, y0 = gradient["start"]
    x1, y1 = gradient["end"]
    x0 = bounds.origin.x + (bounds.size.width * x0)
    y0 = bounds.origin.y + (bounds.size.height * y0)
    x1 = bounds.origin.x + (bounds.size.width * x1)
    y1 = bounds.origin.y + (bounds.size.height * y1)

    colorStop = []
    for stop in gradient.get("colors"):
        color, o = stop
        paletteIndex, a = _find_or_insert_color(color, palette)
        colorStop.append(dict(StopOffset=o, Alpha=a, PaletteIndex=paletteIndex))

    colorLine = dict(Extend="pad", ColorStop=colorStop)

    gradientType = gradient.get("type")
    if gradientType is None:
        # Linear gradient.
        # We set point 2 to point 1 rotated -90 degrees around
        # point 0, since Glyphs gradients don’t represent them.
        paint = dict(
            Format=PaintFormat.PaintLinearGradient,
            ColorLine=colorLine,
            x0=x0,
            y0=y0,
            x1=x1,
            y1=y1,
            x2=x0 + (y1 - y0),
            y2=y0 - (x1 - x0),
        )
    elif gradientType == "circle":
        # Radial gradient.
        # We set radius 0 to 0 and point 1 to point 0, since Glyphs
        # gradients don’t represent them.
        r = _radius(bounds.size, gradient["start"])
        paint = dict(
            Format=PaintFormat.PaintRadialGradient,
            ColorLine=colorLine,
            x0=x0,
            y0=y0,
            x1=x0,
            y1=y0,
            r0=0,
            r1=r,
        )
    else:
        raise NotImplementedError(
            f"Unsupported color layer path gradient in '{layer.parent.name}'"
        )

    return paint


def _to_stroked_paint(attributes, layer, palette, ufo_glyph, ufo):
    try:
        import pathops
    except ImportError as ex:
        raise RuntimeError(
            f"Stroked color layer path in '{layer.parent.name}' requires "
            f"'pathops' module."
        ) from ex

    pos = attributes.get("strokePos", 0)
    if pos:
        raise NotImplementedError(
            f"Unsupported color layer path attribute 'strokePos' "
            f"in '{layer.parent.name}'"
        )

    width = attributes.get("strokeWidth", 1)
    path = pathops.Path()
    ufo_glyph.draw(path.getPen(glyphSet=ufo))
    path.stroke(width, pathops.LineCap.BUTT_CAP, pathops.LineJoin.MITER_JOIN, 4)
    ufo_glyph.clear()
    path.draw(ufo_glyph.getPen())

    color = attributes.get("strokeColor")
    paletteIndex, a = _find_or_insert_color(color, palette)
    return dict(Format=PaintFormat.PaintSolid, Alpha=a, PaletteIndex=paletteIndex)


def _to_component_paint(component):
    paint = dict(Format=PaintFormat.PaintColrGlyph, Glyph=component.name)
    t = Transform(*component.transform)
    if t != Identity:
        if t[:4] == (1, 0, 0, 1):
            return dict(
                Format=PaintFormat.PaintTranslate, Paint=paint, dx=t.dx, dy=t.dy
            )
        else:
            return dict(Format=PaintFormat.PaintTransform, Paint=paint, Transform=t)
    return paint


def _to_ufo_color_layers(builder, ufo, master, layerMapping):
    palette = ([], ufo.lib.get(UFO2FT_COLOR_PALETTES_KEY, [[]])[0])
    for (glyph, masterLayer), layers in builder._color_layers:
        if master.id != masterLayer.associatedMasterId:
            continue

        if glyph.name in layerMapping and isinstance(layerMapping[glyph.name], int):
            raise RuntimeError(
                f"Same glyph can’t have both Color Palette and Color layers: "
                f"{glyph.name}"
            )
        assert glyph.name not in layerMapping

        colorLayers = []
        for i, layer in enumerate(layers):
            if layer.components and layer.attributes:
                colorLayers.append(_to_component_paint(layer.components[0]))
                continue

            if layer.layerId == master.id:
                # This is master layer, we can re-use its UFO glyph.
                layerGlyphName = glyph.name
            else:
                # Not the master layer, create a new UFO glyph for it.
                layerGlyphName = f"{glyph.name}.color{i}"
                ufo_layer = builder.to_ufo_layer(glyph, masterLayer)
                ufo_glyph = ufo_layer.newGlyph(layerGlyphName)
                builder.to_ufo_glyph(
                    ufo_glyph,
                    layer,
                    glyph,
                    do_color_layers=False,
                    is_color_layer_glyph=True,
                )

            attributes = layer.paths[0].attributes if layer.paths else {}
            if "gradient" in attributes:
                gradient = attributes["gradient"]
                paint = _to_gradient_paint(gradient, layer, palette)
            elif "fillColor" in attributes:
                color = attributes["fillColor"]
                paletteIndex, a = _find_or_insert_color(color, palette)
                paint = dict(
                    Format=PaintFormat.PaintSolid, Alpha=a, PaletteIndex=paletteIndex
                )
            elif (
                "strokeColor" in attributes
                or "strokeWidth" in attributes
                or not attributes
            ):
                paint = _to_stroked_paint(attributes, layer, palette, ufo_glyph, ufo)
            else:
                raise NotImplementedError(
                    f"Unsupported color layer path attributes in '{glyph.name}'"
                )

            paintGlyph = dict(
                Format=PaintFormat.PaintGlyph, Glyph=layerGlyphName, Paint=paint
            )
            colorLayers.append(paintGlyph)
        if colorLayers:
            layerMapping[glyph.name] = dict(
                Format=PaintFormat.PaintColrLayers, Layers=colorLayers
            )

    if palette[0]:
        for p in ufo.lib.setdefault(UFO2FT_COLOR_PALETTES_KEY, [[]]):
            p.extend(palette[0])


def to_ufo_color_layers(self, ufo, master):
    layerMapping = {}
    _to_ufo_color_palette_layers(self, master, layerMapping)
    _to_ufo_color_layers(self, ufo, master, layerMapping)
    if layerMapping:
        ufo.lib[UFO2FT_COLOR_LAYERS_KEY] = layerMapping



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import re
from textwrap import dedent
from io import StringIO
from typing import TYPE_CHECKING

from fontTools.feaLib import ast, parser

from glyphsLib.util import PeekableIterator
from .constants import (
    GLYPHLIB_PREFIX,
    ANONYMOUS_FEATURE_PREFIX_NAME,
    ORIGINAL_FEATURE_CODE_KEY,
    ORIGINAL_CATEGORY_KEY,
    LANGUAGE_MAPPING,
    REVERSE_LANGUAGE_MAPPING,
    INSERT_FEATURE_MARKER_RE,
    INSERT_FEATURE_MARKER_COMMENT,
)
from .tokens import TokenExpander, PassThruExpander

if TYPE_CHECKING:
    from ufoLib2 import Font

    from ..classes import GSFont, GSFontMaster
    from . import UFOBuilder


def autostr(automatic):
    return "# automatic\n" if automatic else ""


def to_ufo_master_features(self, ufo, master):
    # Recover the original feature code if it was stored in the user data
    original = master.userData[ORIGINAL_FEATURE_CODE_KEY]
    if original is not None:
        ufo.features.text = original
    else:
        ufo.features.text = _to_ufo_features(
            self.font,
            ufo,
            generate_GDEF=self.generate_GDEF,
            master=master,
            expand_includes=self.expand_includes,
            minimal=self.minimal,
        )


def _to_name_langID(language):
    if language not in LANGUAGE_MAPPING:
        raise ValueError(f"Unknown name language: {language}")
    return LANGUAGE_MAPPING[language]


def _to_glyphs_language(langID):
    if langID not in REVERSE_LANGUAGE_MAPPING:
        raise ValueError(f"Unknown name langID: {langID}")
    return REVERSE_LANGUAGE_MAPPING[langID]


def _is_manual_kern_feature(feature):
    """Return true if the feature is a manually written 'kern' features."""
    return feature.name == "kern" and not feature.automatic


def _to_ufo_features(  # noqa: C901
    font: GSFont,
    ufo: Font | None = None,
    generate_GDEF: bool = False,
    master: GSFontMaster | None = None,
    expand_includes: bool = False,
    minimal: bool = False,
) -> str:
    """Convert GSFont features, including prefixes and classes, to UFO.

    Optionally, build a GDEF table definiton, excluding 'skip_export_glyphs'.
    """
    if not master:
        expander = PassThruExpander()
    else:
        expander = TokenExpander(font, master)

    prefixes = []
    for prefix in font.featurePrefixes:
        if prefix.disabled and minimal:
            continue
        strings = []
        if prefix.name != ANONYMOUS_FEATURE_PREFIX_NAME:
            strings.append("# Prefix: %s\n" % prefix.name)
        strings.append(autostr(prefix.automatic))
        strings.append(expander.expand(prefix.code))
        prefixes.append("".join(strings))

    prefix_str = "\n\n".join(prefixes)

    class_defs = []
    for class_ in font.classes:
        if class_.disabled and minimal:
            continue
        prefix = "@" if not class_.name.startswith("@") else ""
        name = prefix + class_.name
        class_defs.append(
            "{}{} = [ {}\n];".format(
                autostr(class_.automatic), name, expander.expand(class_.code)
            )
        )
    class_str = "\n\n".join(class_defs)

    feature_defs = []
    for feature in font.features:
        if feature.disabled and minimal:
            continue
        code = expander.expand(feature.code)
        lines = ["feature %s {" % feature.name]
        notes = feature.notes
        feature_names = None
        if font.format_version == 2 and notes:
            m = re.search("(featureNames {.+};)", notes, flags=re.DOTALL)
            if m:
                name = m.groups()[0]
                # Remove the name from the note
                notes = notes.replace(name, "").strip()
                feature_names = name.splitlines()
            else:
                m = re.search(r"^(Name: (.+))", notes)
                if m:
                    line, name = m.groups()
                    # Remove the name from the note
                    notes = notes.replace(line, "").strip()
                    # Replace special chars backslash and doublequote for AFDKO syntax
                    name = name.replace("\\", r"\005c").replace('"', r"\0022")
                    feature_names = ["featureNames {", f'  name "{name}";', "};"]
        elif font.format_version == 3 and feature.labels:
            feature_names = []
            for label in feature.labels:
                langID = _to_name_langID(label["language"])
                name = label["value"]
                if name == "":
                    continue
                name = name.replace("\\", r"\005c").replace('"', r"\0022")
                if langID is None:
                    feature_names.append(f'  name "{name}";')
                else:
                    feature_names.append(f'  name 3 1 0x{langID:X} "{name}";')
            if feature_names:
                feature_names.insert(0, "featureNames {")
                feature_names.append("};")
        if notes:
            lines.append("# notes:")
            lines.extend("# " + line for line in notes.splitlines())
        if feature_names:
            lines.extend(feature_names)
        if feature.automatic:
            lines.append("# automatic")
        if feature.disabled:
            lines.append("# disabled")
            lines.extend("#" + line for line in code.splitlines())
        else:
            lines.append(code)

        # Manual kern features in glyphs also have the automatic code added after them
        # We make sure it gets added with an "Automatic Code..." marker if it doesn't
        # already have one.
        if _is_manual_kern_feature(feature) and not re.search(
            INSERT_FEATURE_MARKER_RE, code
        ):
            lines.append(INSERT_FEATURE_MARKER_COMMENT)

        lines.append("} %s;" % feature.name)
        feature_defs.append("\n".join(lines))
    fea_str = "\n\n".join(feature_defs)

    if generate_GDEF:
        assert ufo is not None
        regenerate_opentype_categories(font, ufo)

    full_text = "\n\n".join(filter(None, [class_str, prefix_str, fea_str])) + "\n"
    full_text = full_text if full_text.strip() else ""

    if not full_text or not expand_includes:
        return full_text

    # use feaLib Parser to resolve include statements relative to the GSFont
    # fontpath, and inline them in the output features text.
    feature_file = StringIO(full_text)
    include_dir = os.path.dirname(font.filepath) if font.filepath else None
    fea_parser = parser.Parser(
        feature_file,
        glyphNames={glyph.name for glyph in font.glyphs},
        includeDir=include_dir,
        followIncludes=expand_includes,
    )
    doc = fea_parser.parse()
    return doc.asFea()


def _build_public_opentype_categories(ufo: Font) -> dict[str, str]:
    """Returns a dictionary mapping glyph names to GDEF categories.

    Does not handle ligature carets. A GDEF table with both categories and
    ligature carets is generated by ufo2ft's GdefFeatureWriter at compile time,
    using the categories dict and glyph data.

    Determining the categories requires anchor propagation or user care to work
    as expected, as Glyphs.app also looks at anchors for classification:

    * Base: any glyph that has an attaching anchor (such as "top"; "_top" does
      not count) and is neither classified as Ligature nor Mark using the
      definitions below;
    * Ligature: if subCategory is "Ligature" and the glyph has at least one
      attaching anchor;
    * Mark: if category is "Mark" and subCategory is either "Nonspacing" or
      "Spacing Combining";
    * Compound: never assigned by Glyphs.app.

    See:

    * https://github.com/googlefonts/glyphsLib/issues/85
    * https://github.com/googlefonts/glyphsLib/pull/100#issuecomment-275430289
    """
    from glyphsLib import glyphdata

    categories: dict[str, str] = {}
    category_key = GLYPHLIB_PREFIX + "category"
    subCategory_key = GLYPHLIB_PREFIX + "subCategory"

    # NOTE: We can generate the category even for glyphs that are not exported,
    # because entries don't have to exist in the final fonts.
    for glyph in ufo:
        glyph_name = glyph.name
        assert glyph_name is not None

        has_attaching_anchor = False
        for anchor in glyph.anchors:
            name = anchor.name
            if name and not name.startswith("_"):
                has_attaching_anchor = True

        # First check glyph.lib for category/subCategory overrides. Otherwise,
        # use global values from GlyphData.
        glyphinfo = glyphdata.get_glyph(
            glyph_name, unicodes=[f"{c:04X}" for c in glyph.unicodes]
        )
        category = glyph.lib.get(category_key) or glyphinfo.category
        subCategory = glyph.lib.get(subCategory_key) or glyphinfo.subCategory

        if subCategory == "Ligature" and has_attaching_anchor:
            categories[glyph_name] = "ligature"
        elif category == "Mark" and (
            subCategory == "Nonspacing" or subCategory == "Spacing Combining"
        ):
            categories[glyph_name] = "mark"
        elif has_attaching_anchor:
            categories[glyph_name] = "base"

    return categories


def regenerate_gdef(self: UFOBuilder) -> None:
    for source in self._sources.values():
        regenerate_opentype_categories(self.font, source.font)


def regenerate_opentype_categories(font: GSFont, ufo: Font) -> None:
    categories = _build_public_opentype_categories(ufo)

    # Prefer already stored categories for round-tripping. This will provide
    # newly guessed categories only for new glyphs. The data is stored
    # GSFont-wide to capture bracket glyphs that we create for UFOs and fold
    # when going back.
    roundtripping_categories = font.userData[ORIGINAL_CATEGORY_KEY]
    if roundtripping_categories is not None:
        categories.update(roundtripping_categories)

    if categories:
        ufo.lib["public.openTypeCategories"] = categories


def _replace_block(kind, tag, repl, features):
    if not repl.endswith("\n"):
        repl += "\n"
    return re.sub(
        rf"(?<=^{kind} {tag} {{\n)(.*?)(?=^}} {tag};$)",
        repl,
        features,
        count=1,
        flags=re.DOTALL | re.MULTILINE,
    )


def replace_feature(tag, repl, features):
    return _replace_block("feature", tag, repl, features)


def replace_table(tag, repl, features):
    return _replace_block("table", tag, repl, features)


def replace_prefixes(repl_map, features_text, glyph_names=None):
    """Replace all '# Prefix: NAME' sections in features.

    Args:
        repl_map: Dict[str, str]: dictionary keyed by prefix name containing
            feature code snippets to be replaced.
        features_text: str: feature text to be parsed.
        glyph_names: Optional[Sequence[str]]: list of valid glyph names, used
            by feaLib Parser to distinguish glyph name tokens containing '-' from
            glyph ranges such as 'a-z'.

    Returns:
        str: new feature text with replaced prefix paragraphs.
    """
    from glyphsLib.classes import GSFont

    temp_font = GSFont()
    _to_glyphs_features(temp_font, features_text, glyph_names=glyph_names)

    for prefix in temp_font.featurePrefixes:
        if prefix.name in repl_map:
            prefix.code = repl_map[prefix.name]

    return _to_ufo_features(temp_font)


# UFO to Glyphs


def to_glyphs_features(self):
    if not self.designspace.sources:
        # Needs at least one UFO
        return

    # Handle differing feature files between input UFOs
    # For now: switch to very simple strategy if there is any difference
    # TODO: (jany) later, use a merge-as-we-go strategy where all discovered
    #   features go into the GSFont's features, and custom parameters are used
    #   to disable features on masters that didn't have them originally.
    if _features_are_different_across_ufos(self):
        if self.minimize_ufo_diffs:
            self.logger.warning(
                "Feature files are different across UFOs. The produced Glyphs "
                "file will have no editable features."
            )
            # Do all UFOs, not only the first one
            _to_glyphs_features_basic(self)
            return
        self.logger.warning(
            "Feature files are different across UFOs. The produced Glyphs "
            "file will reflect only the features of the first UFO."
        )

    # Split the feature file of the first UFO into GSFeatures
    ufo = self.designspace.sources[0].font
    if ufo.features.text is None:
        return
    include_dir = None
    if self.expand_includes and ufo.path:
        include_dir = os.path.dirname(os.path.normpath(ufo.path))
    _to_glyphs_features(
        self.font,
        ufo.features.text,
        glyph_names=ufo.keys(),
        glyphs_module=self.glyphs_module,
        include_dir=include_dir,
        expand_includes=self.expand_includes,
    )

    # Store GDEF category data GSFont-wide to capture bracket glyphs that we
    # create for UFOs and fold when going back.
    opentype_categories = ufo.lib.get("public.openTypeCategories")
    if opentype_categories is not None:
        self.font.userData[ORIGINAL_CATEGORY_KEY] = opentype_categories


def _to_glyphs_features(
    font,
    features_text,
    glyph_names=None,
    glyphs_module=None,
    include_dir=None,
    expand_includes=False,
):
    """Import features text in GSFont, split into prefixes, features and classes.

    Args:
        font: GSFont
        feature_text: str
        glyph_names: Optional[Sequence[str]]
        glyphs_module: Optional[Any]
        include_dir: Optional[str]
        expand_includes: bool
    """
    document = FeaDocument(
        features_text,
        glyph_names,
        include_dir=include_dir,
        expand_includes=expand_includes,
    )
    processor = FeatureFileProcessor(document, glyphs_module)
    processor.to_glyphs(font)


def _features_are_different_across_ufos(self):
    # FIXME: requires that features are in the same order in all feature files;
    #   the only allowed differences are whitespace
    # Construct iterator over full UFOs, layer sources do not have a font object set.
    full_sources = (s for s in self.designspace.sources if not s.layerName)
    reference = next(full_sources).font.features.text or ""
    reference = _normalize_whitespace(reference)
    for source in full_sources:
        other = _normalize_whitespace(source.font.features.text or "")
        if reference != other:
            return True
    return False


def _normalize_whitespace(text):
    # FIXME: does not take into account "significant" whitespace like
    # whitespace in a UI string
    return re.sub(r"\s+", " ", text)


def _to_glyphs_features_basic(self):
    prefix = self.glyphs_module.GSFeaturePrefix()
    prefix.name = "WARNING"
    prefix.code = dedent(
        """\
        # Do not use Glyphs to edit features.
        #
        # This Glyphs file was made from several UFOs that had different
        # features. As a result, the features are not editable in Glyphs and
        # the original features will be restored when you go back to UFOs.
    """
    )
    self.font.featurePrefixes.append(prefix)
    for master_id, source in self._sources.items():
        master = self.font.masters[master_id]
        master.userData[ORIGINAL_FEATURE_CODE_KEY] = source.font.features.text


class FeaDocument:
    """Parse the string of a fea code into statements."""

    def __init__(self, text, glyph_set=None, include_dir=None, expand_includes=False):
        feature_file = StringIO(text)
        glyph_names = glyph_set if glyph_set is not None else ()
        parser_ = parser.Parser(
            feature_file,
            glyphNames=glyph_names,
            includeDir=include_dir,
            followIncludes=expand_includes,
        )
        if expand_includes:
            # if we expand includes, we need to reparse the whole file with the
            # new content to get the updated locations
            text = parser_.parse().asFea()
            parser_ = parser.Parser(StringIO(text), glyphNames=glyph_names)
        self._doc = parser_.parse()
        self.statements = self._doc.statements
        self._lines = text.splitlines(True)  # keepends=True
        self._build_end_locations()

    def text(self, statements):
        """Recover the original fea code of the given statements from the
        given block.
        """
        return "".join(self._statement_text(st) for st in statements)

    def _statement_text(self, statement):
        _, begin_line, begin_char = statement.location
        _, end_line, end_char = statement.end_location
        lines = self._lines[begin_line - 1 : end_line]
        if lines:
            # In case it's the same line, we need to trim the end first
            lines[-1] = lines[-1][:end_char]
            lines[0] = lines[0][begin_char - 1 :]
        return "".join(lines)

    def _build_end_locations(self):
        # The statements in the ast only have their start location, but we also
        # need the end location to find the text in between.
        # FIXME: (jany) maybe feaLib could provide that?
        # Add a fake statement at the end, it's the only one that won't get
        # a proper end_location, but its presence will help compute the
        # end_location of the real last statement(s).
        self._lines.append("#")  # Line corresponding to the fake statement
        fake_location = (None, len(self._lines), 1)
        self._doc.statements.append(
            ast.Comment(text="Sentinel", location=fake_location)
        )
        self._build_end_locations_rec(self._doc)
        # Remove the fake last statement
        self._lines.pop()
        self._doc.statements.pop()

    def _build_end_locations_rec(self, block):
        # To get the end location, we do a depth-first exploration of the ast:
        # When a new statement starts, it means that the previous one ended.
        # When a new statement starts outside of the current block, we must
        # remove the "end-of-block" string from the previous inner statement.
        previous = None
        previous_in_block = None
        for st in block.statements:
            if hasattr(st, "statements"):
                self._build_end_locations_rec(st)
            if previous is not None:
                _, line, char = st.location
                line, char = self._previous_char(line, char)
                previous.end_location = (None, line, char)
            if previous_in_block is not None:
                previous_in_block.end_location = self._in_block_end_location(previous)
                previous_in_block = None
            previous = st
            if hasattr(st, "statements"):
                previous_in_block = st.statements[-1] if st.statements else None

    WHITESPACE_RE = re.compile("\\s")
    WHITESPACE_OR_NAME_RE = re.compile("\\w|\\s")

    def _previous_char(self, line, char):
        char -= 1
        while char == 0:
            line -= 1
            char = len(self._lines[line - 1])
        return line, char

    def _in_block_end_location(self, block):
        _, line, char = block.end_location

        def current_char(line, char):
            return self._lines[line - 1][char - 1]

        # Find the semicolon
        while current_char(line, char) != ";":
            assert self.WHITESPACE_RE.match(current_char(line, char))
            line, char = self._previous_char(line, char)
        # Skip it
        line, char = self._previous_char(line, char)
        # Skip the whitespace and table/feature name
        while self.WHITESPACE_OR_NAME_RE.match(current_char(line, char)):
            line, char = self._previous_char(line, char)
        # It should be the closing bracket
        assert current_char(line, char) == "}"
        # Skip it and we're done
        line, char = self._previous_char(line, char)

        return None, line, char


class FeatureFileProcessor:
    """Put fea statements into the correct fields of a GSFont."""

    def __init__(self, doc, glyphs_module=None):
        self.doc = doc
        if glyphs_module is None:
            from glyphsLib import classes as glyphs_module
        self.glyphs_module = glyphs_module
        self.statements = PeekableIterator(doc.statements)
        self._font = None

    def to_glyphs(self, font):
        self._font = font
        self._process_file()

    PREFIX_RE = re.compile("^# Prefix: (.*)$")
    AUTOMATIC_RE = re.compile("^# automatic$")
    DISABLED_RE = re.compile("^# disabled$")
    NOTES_RE = re.compile("^# notes:$")

    def _process_file(self):
        unhandled_root_elements = []
        while self.statements.has_next():
            if (
                self._process_prefix()
                or self._process_glyph_class_definition()
                or self._process_feature_block()
                or self._process_gdef_table_block()
            ):
                # Flush any unhandled root elements into an anonymous prefix
                if unhandled_root_elements:
                    prefix = self.glyphs_module.GSFeaturePrefix()
                    prefix.name = ANONYMOUS_FEATURE_PREFIX_NAME
                    prefix.code = self._rstrip_newlines(
                        self.doc.text(unhandled_root_elements)
                    )
                    self._font.featurePrefixes.append(prefix)
                    del unhandled_root_elements[:]  # list.clear() in Python 3.
            else:
                # FIXME: (jany) Maybe print warning about unhandled fea block?
                unhandled_root_elements.append(next(self.statements))
        # Flush any unhandled root elements into an anonymous prefix
        if unhandled_root_elements:
            prefix = self.glyphs_module.GSFeaturePrefix()
            prefix.name = ANONYMOUS_FEATURE_PREFIX_NAME
            prefix.code = self._rstrip_newlines(self.doc.text(unhandled_root_elements))
            self._font.featurePrefixes.append(prefix)

    def _process_prefix(self):
        st = self.statements.peek()
        if not isinstance(st, ast.Comment):
            return False
        match = self.PREFIX_RE.match(st.text)
        if not match:
            return False
        next(self.statements)

        # Consume statements that are part of the feature prefix
        prefix_statements = []
        while self.statements.has_next():
            st = self.statements.peek()
            # Don't consume statements that are treated specially
            if isinstance(
                st, (ast.GlyphClassDefinition, ast.FeatureBlock, ast.TableBlock)
            ):
                break
            # Don't comsume a comment if it is the start of another prefix...
            if isinstance(st, ast.Comment):
                if self.PREFIX_RE.match(st.text):
                    break
                # ...or if it is the "automatic" comment just before a class
                if self.statements.has_next(1):
                    next_st = self.statements.peek(1)
                    if self.AUTOMATIC_RE.match(st.text) and isinstance(
                        next_st, ast.GlyphClassDefinition
                    ):
                        break
            prefix_statements.append(next(self.statements))

        prefix = self.glyphs_module.GSFeaturePrefix()
        prefix.name = match.group(1)
        automatic, prefix_statements = self._pop_comment(
            prefix_statements, self.AUTOMATIC_RE
        )
        prefix.automatic = bool(automatic)
        prefix.code = self._rstrip_newlines(self.doc.text(prefix_statements), 2)
        self._font.featurePrefixes.append(prefix)
        return True

    def _process_glyph_class_definition(self):
        automatic = False
        st = self.statements.peek()
        if isinstance(st, ast.Comment):
            if self.AUTOMATIC_RE.match(st.text):
                automatic = True
                st = self.statements.peek(1)
            else:
                return False
        if not isinstance(st, ast.GlyphClassDefinition):
            return False
        if automatic:
            next(self.statements)
        next(self.statements)
        glyph_class = self.glyphs_module.GSClass()
        glyph_class.name = st.name
        # Call st.glyphs.asFea() because it updates the 'original' field
        # However, we don't use the result of `asFea` because it expands
        # classes in a strange way
        # FIXME: (jany) maybe open an issue if feaLib?
        st.glyphs.asFea()
        elements = []
        try:
            if st.glyphs.original:
                for glyph in st.glyphs.original:
                    try:
                        # Class name (MarkClassName object)
                        elements.append("@" + glyph.markClass.name)
                    except AttributeError:
                        try:
                            # Class name (GlyphClassName object)
                            elements.append("@" + glyph.glyphclass.name)
                        except AttributeError:
                            try:
                                # Class name (GlyphClassDefinition object)
                                # FIXME: (jany) why not always the same type?
                                elements.append("@" + glyph.name)
                            except AttributeError:
                                # Glyph name
                                elements.append(glyph)
            else:
                elements = st.glyphSet()
        except AttributeError:
            # Single class
            try:
                # Class name (MarkClassName object)
                elements.append("@" + st.glyphs.markClass.name)
            except AttributeError:
                # Class name (GlyphClassName object)
                elements.append("@" + st.glyphs.glyphclass.name)
        glyph_class.code = " ".join(elements)
        glyph_class.automatic = bool(automatic)
        self._font.classes.append(glyph_class)
        return True

    def _process_feature_block(self):
        st = self.statements.peek()
        if not isinstance(st, ast.FeatureBlock):
            return False
        next(self.statements)
        contents = st.statements
        automatic, contents = self._pop_comment(contents, self.AUTOMATIC_RE)
        disabled, disabled_text, contents = self._pop_comment_block(
            contents, self.DISABLED_RE
        )
        notes, notes_text, contents = self._pop_comment_block(contents, self.NOTES_RE)
        feature = self.glyphs_module.GSFeature()
        feature.name = st.name
        feature.automatic = bool(automatic)
        if feature.automatic:
            # See if there is a feature names block in the code that should be
            # written to the notes.
            for i, statement in enumerate(contents):
                if (
                    isinstance(statement, ast.NestedBlock)
                    and statement.block_name == "featureNames"
                ):
                    feature_names = contents[i]
                    if feature_names.statements:
                        # If there is only one name has default platformID,
                        # platEncID and langID, write it using the simple
                        # syntax. Otherwise write out the full featureNames
                        # statement.
                        if self._font.format_version == 2:
                            name = feature_names.statements[0]
                            if (
                                len(feature_names.statements) == 1
                                and name.platformID == 3
                                and name.platEncID == 1
                                and name.langID == 0x409
                            ):
                                name_text = f"Name: {name.string}"
                            else:
                                name_text = str(feature_names)
                            notes_text = name_text + "\n" + notes_text
                            notes = True
                            contents.pop(i)
                        elif self._font.format_version == 3:
                            labels = []
                            for name in feature_names.statements:
                                if name.platformID == 3 and name.platEncID == 1:
                                    language = _to_glyphs_language(name.langID)
                                    labels.append(
                                        dict(language=language, value=name.string)
                                    )
                            if len(labels) == len(feature_names.statements):
                                feature.labels = labels
                                contents.pop(i)
                    break
        if notes:
            feature.notes = notes_text
        if disabled:
            feature.code = disabled_text
            feature.disabled = True
            # FIXME: (jany) check that the user has not added more new code
            #    after the disabled comment. Maybe start by checking whether
            #    the block is only made of comments
        else:
            feature.code = self._rstrip_newlines(self.doc.text(contents))
        self._font.features.append(feature)
        return True

    def _process_gdef_table_block(self):
        st = self.statements.peek()
        if not isinstance(st, ast.TableBlock) or st.name != "GDEF":
            return False
        # TODO: read an existing GDEF table and do something with it?
        # For now, this function returns False to say that it has not handled
        # the GDEF table, so it will be stored in Glyphs as a prefix with other
        # "unhandled root elements".
        return False

    def _pop_comment(self, statements, comment_re):
        """Look for the comment that matches the given regex.
        If it matches, return the regex match object and list of statements
        without the special one.
        """
        res = []
        match = None
        for st in statements:
            if match or not isinstance(st, ast.Comment):
                res.append(st)
                continue
            match = comment_re.match(st.text)
            if not match:
                res.append(st)
        return match, res

    def _pop_comment_block(self, statements, header_re):
        """Look for a series of comments that start with one that matches the
        regex. If the first comment is found, all subsequent comments are
        popped from statements, concatenated and dedented and returned.
        """
        res = []
        comments = []
        match = None
        st_iter = iter(statements)
        # Look for the header
        for st in st_iter:
            if isinstance(st, ast.Comment):
                match = header_re.match(st.text)
                if match:
                    # Drop this comment an move on to consuming the block
                    break
                else:
                    res.append(st)
            else:
                res.append(st)
        # Consume consecutive comments
        for st in st_iter:
            if isinstance(st, ast.Comment):
                comments.append(st)
            else:
                # The block is over, keep the rest of the statements
                res.append(st)
                break
        # Keep the rest of the statements
        res.extend(list(st_iter))
        # Inside the comment block, drop the pound sign and any common indent
        return match, dedent("".join(c.text[1:] + "\n" for c in comments)), res

    # Strip up to the given number of newlines from the right end of the string
    def _rstrip_newlines(self, string, number=1):
        for _ in range(number):
            if string and string[-1] == "\n":
                string = string[:-1]
        return string
